import logging
import os
import subprocess
from enum import Enum
from functools import lru_cache
from pathlib import Path
from shutil import which
from typing import Dict, List, Optional, Union

from ggshield.utils.os import getenv_int


EMPTY_SHA = "0000000000000000000000000000000000000000"
EMPTY_TREE = "4b825dc642cb6eb9a060e54bf8d69288fbee4904"

COMMAND_TIMEOUT = getenv_int("GG_GIT_TIMEOUT", 45)

logger = logging.getLogger(__name__)


class GitError(Exception):
    pass


class InvalidGitRefError(GitError):
    """
    Raised when the git reference does not exist
    """

    def __init__(self, ref: str):
        super().__init__(f"Not a git reference: {ref}.")


class NotAGitDirectory(GitError):
    def __init__(self):
        super().__init__("Not a git directory.")


class GitExecutableNotFound(GitError):
    pass


class GitCommandTimeoutExpired(GitError):
    pass


class Filemode(Enum):
    """
    Enum class for git filemode.
    """

    MODIFY = "modified file"
    DELETE = "deleted file"
    NEW = "new file"
    RENAME = "renamed file"
    FILE = "file"
    UNKNOWN = "unknown"


@lru_cache(None)
def _get_git_path() -> str:
    git_path = which("git")

    if git_path is None:
        raise GitExecutableNotFound("unable to find git executable in PATH/PATHEXT")

    # lower()ing these would provide additional coverage on case-
    # insensitive filesystems but detection is problematic
    git_path = os.path.abspath(git_path)
    cwd = os.getcwd()
    path_env = [
        os.path.abspath(p) for p in os.environ.get("PATH", "").split(os.pathsep)
    ]

    # git was found - ignore git in cwd if cwd not in PATH
    if cwd == os.path.dirname(git_path) and cwd not in path_env:
        raise GitExecutableNotFound("rejecting git executable in CWD not in PATH")

    logger.debug("Found git at %s", git_path)
    return git_path


@lru_cache(None)
def _git_rev_parse_absolute(option: str, wd_absolute: Path) -> Optional[str]:
    """
    Helper function for `_git_rev_parse` to only cache on absolute paths.
    """
    try:
        return git(["rev-parse", option], cwd=wd_absolute)
    except subprocess.CalledProcessError:
        return None


def _git_rev_parse(option: str, wd: Path) -> Optional[str]:
    return _git_rev_parse_absolute(option=option, wd_absolute=wd.resolve())


def is_git_dir(wd: Union[str, Path]) -> bool:
    return _git_rev_parse("--git-dir", Path(wd)) is not None


def is_git_working_tree(wd: Union[str, Path]) -> bool:
    return _git_rev_parse("--show-toplevel", Path(wd)) is not None


def get_git_root(wd: Optional[Union[str, Path]] = None) -> Path:
    """
    Fetches the root of the git repo.
    This corresponds to the root directory in the case of a working tree,
    or the `.git/` directory in the case of a quarantine during pre-receive.

    :param wd: working directory, defaults to None
    :return: absolute path to the git root, as a string.
    """
    if wd is None:
        wd = Path.cwd()
    else:
        wd = Path(wd)
    check_git_dir(wd)
    top_level = _git_rev_parse(option="--show-toplevel", wd=wd)
    if top_level is not None:
        return Path(top_level)
    root = _git_rev_parse(option="--git-dir", wd=wd)
    if root is None:
        raise NotAGitDirectory()
    return Path(root).resolve()


def check_git_dir(wd: Optional[Union[str, Path]] = None) -> None:
    """Check if folder is git directory."""
    if wd is None:
        wd = Path.cwd()
    if not is_git_dir(wd):
        raise NotAGitDirectory()


def git(
    command: List[str],
    timeout: int = COMMAND_TIMEOUT,
    check: bool = True,
    cwd: Optional[Union[str, Path]] = None,
) -> str:
    """Calls git with the given arguments, returns stdout as a string"""
    env = os.environ.copy()
    env["LANG"] = "C"

    if cwd is None:
        cwd = Path.cwd()

    try:
        logger.debug("command=%s timeout=%d", command, timeout)
        result = subprocess.run(
            [_get_git_path()] + command,
            check=check,
            capture_output=True,
            timeout=timeout,
            env=env,
            cwd=str(cwd),
        )
        if result.stderr:
            logger.debug("stderr=%s", result.stderr.decode("utf-8", errors="ignore"))
        return result.stdout.decode("utf-8", errors="ignore").rstrip()
    except subprocess.CalledProcessError as e:
        if "detected dubious ownership in repository" in e.stderr.decode(
            "utf-8", errors="ignore"
        ):
            raise GitError(
                "Git command failed because of a dubious ownership in repository.\n"
                "If you still want to run ggshield, make sure you mark "
                "the current repository as safe for git with:\n"
                "   git config --global --add safe.directory <YOUR_REPO>"
            )
        raise e
    except subprocess.TimeoutExpired:
        raise GitCommandTimeoutExpired(
            'Command "{}" timed out'.format(" ".join(command))
        )


def git_ls(wd: Optional[Union[str, Path]] = None) -> List[str]:
    cmd = ["ls-files", "--recurse-submodules"]
    return git(cmd, timeout=600, cwd=wd).split("\n")


def is_valid_git_commit_ref(ref: str, wd: Optional[Union[str, Path]] = None) -> bool:
    """
    Check if a reference is valid and can be resolved to a commit
    """
    if not wd:
        wd = Path.cwd()

    ref += "^{commit}"
    cmd = ["cat-file", "-e", ref]

    try:
        git(cmd, cwd=wd)
    except subprocess.CalledProcessError:
        return False

    return True


def check_git_ref(ref: str, wd: Optional[Union[str, Path]] = None) -> None:
    """Check if folder is a git repository and ref is a git reference."""
    if wd is None:
        wd = Path.cwd()
    check_git_dir(wd)

    if not is_valid_git_commit_ref(ref=ref, wd=wd):
        raise InvalidGitRefError(ref)


def get_list_commit_SHA(
    commit_range: str, max_count: Optional[int] = None
) -> List[str]:
    """
    Retrieve the list of commit SHA from a range.
    :param commit_range: A range of commits (ORIGIN...HEAD)
    :param max_count: If set, limits the number of SHA returned to this amount. This
    returns the *end* of the list, so max_count=3 returns [HEAD~2, HEAD~1, HEAD].
    """

    cmd = ["rev-list", "--reverse", *commit_range.split()]
    if max_count is not None:
        cmd.extend(["--max-count", str(max_count)])
    # Makes rev-list print "bad revision" instead of telling the range is ambiguous
    cmd.append("--")

    try:
        commit_list = git(cmd).split("\n")
    except subprocess.CalledProcessError as e:
        if b"bad revision" in e.stderr and "~1.." in commit_range:
            # We got asked to list commits for A~1...B. If A~1 does not exist, but A
            # does, then return A and its descendants until B.
            a_ref, remaining = commit_range.split("~1", maxsplit=1)
            if not is_valid_git_commit_ref(f"{a_ref}~1") and is_valid_git_commit_ref(
                a_ref
            ):
                commit_range = a_ref + remaining
                return [a_ref] + get_list_commit_SHA(commit_range)
        return []

    if "" in commit_list:
        commit_list.remove("")
        # only happens when git rev-list doesn't error
        # but returns an empty range, example git rev-list HEAD...

    return commit_list


def get_last_commit_sha_of_branch(branch_name: str) -> Optional[str]:
    """
    Returns the last commit sha of the given branch, or None
    if no commit could be found
    """
    # The branch is not directly available in CI env
    # We need to get commits through remotes
    last_target_commit = get_list_commit_SHA(branch_name, max_count=1)

    # Unable to find a commit on this branch
    # Consider it empty
    if not last_target_commit:
        return None

    return last_target_commit[0]


def get_filepaths_from_ref(
    ref: str, wd: Optional[Union[str, Path]] = None
) -> List[Path]:
    """
    Fetches a list of all file paths indexed at a given reference in a git repository.
    :param ref: git reference, like a commit SHA, a relative reference like HEAD~1, ...
    :param wd: string path to the git repository. Defaults to current directory
    """
    if not wd:
        wd = Path.cwd()

    check_git_ref(ref, wd)

    filepaths = git(
        ["ls-tree", "--name-only", "--full-name", "-r", ref], cwd=wd
    ).splitlines()
    return [Path(path_str) for path_str in filepaths]


def get_staged_filepaths(wd: Optional[Union[str, Path]] = None) -> List[Path]:
    """
    Fetches a list of all file paths at the index in a git repository.
    :param wd: string path to the git repository. Defaults to current directory
    """
    if not wd:
        wd = Path.cwd()

    filepaths = git(["ls-files", "--full-name", "-c"], cwd=wd).splitlines()
    return [Path(path_str) for path_str in filepaths]


def get_diff_files_status(
    ref: str,
    staged: bool = False,
    similarity: int = 100,
    wd: Optional[Union[str, Path]] = None,
    current_ref: Optional[str] = None,
) -> Dict[Path, Filemode]:
    """
    Fetches the statuses of modified files since a given ref.
    For more details on file statuses, see:
    https://git-scm.com/docs/git-diff#Documentation/git-diff.txt---diff-filterACDMRTUXB82308203
    """

    # Input validation

    assert 0 <= similarity <= 100

    if current_ref is None:
        current_ref = "HEAD"

    if not wd:
        wd = Path.cwd()

    check_git_ref(wd=wd, ref=ref)

    if not staged:
        check_git_ref(wd=wd, ref=current_ref)

    if ref == "HEAD" and not staged:
        return dict()

    def parse_name_status_patch(patch: str) -> Dict[Path, Filemode]:
        status_to_filemode = {
            "A": Filemode.NEW,
            "D": Filemode.DELETE,
            "M": Filemode.MODIFY,
            "T": Filemode.MODIFY,
            "R": Filemode.RENAME,
        }

        split_patch = patch.split("\0")
        chunks = (split_patch[i : i + 2] for i in range(0, len(split_patch) - 2, 2))

        return {
            Path(path): status_to_filemode.get(mode, Filemode.UNKNOWN)
            for mode, path in chunks
        }

    is_working_tree = is_git_working_tree(wd)
    cmd = [
        "diff" if is_working_tree else "diff-tree",
        f"-M{similarity}%",
        "--name-status",
        "--raw",
        "-z",
        "--patch",
        "--diff-filter=ADMTR",
    ]

    if staged and is_working_tree:
        cmd.append("--staged")

    if not is_working_tree:
        cmd.append(current_ref)
    cmd.append(ref)

    patch = git(cmd, cwd=wd)

    return parse_name_status_patch(patch)


@lru_cache(None)
def read_git_file(ref: str, path: Path, wd: Optional[Union[str, Path]] = None) -> str:
    # Use as_posix to handle git and Windows
    return git(["show", f"{ref}:{path.as_posix()}"], cwd=wd)



from typing import Any

import click

from ggshield.cmd.iac.scan import iac_scan_group
from ggshield.cmd.utils.common_options import add_common_options


@click.group(commands={"scan": iac_scan_group})
@add_common_options()
def iac_group(**kwargs: Any) -> None:
    """Commands to work with Infra as Code."""



from abc import ABC, abstractproperty
from enum import Enum
from typing import Optional, Union

from pygitguardian.sca_models import SCAScanAllOutput, SCAScanDiffOutput


SCAScanResult = Union[SCAScanAllOutput, SCAScanDiffOutput]


class SCAVulnerabilityCollectionType(Enum):
    UNKNOWN = "unknown"
    DIRECTORY = "sca_directory"
    DIFF = "sca_diff"


class SCAScanVulnerabilityCollection(ABC):
    """
    A Collection of SCA vulnerabilities
    """

    type = SCAVulnerabilityCollectionType.UNKNOWN

    def __init__(
        self,
        id: str,
        # Can be None if the scan failed
        result: Optional[SCAScanResult],
    ):
        self.id = id
        self.result = result

    @abstractproperty
    def has_results(self) -> bool:
        """
        Whether the scan found problems
        """
        return self.result is not None


class SCAScanAllVulnerabilityCollection(SCAScanVulnerabilityCollection):
    type = SCAVulnerabilityCollectionType.DIRECTORY
    result: Optional[SCAScanAllOutput]

    @property
    def has_results(self) -> bool:
        """
        Checks if at least one package in a location has one vulnerability
        """
        return self.result is not None and any(
            len(package.vulns) > 0
            for location in self.result.found_package_vulns
            for package in location.package_vulns
        )


class SCAScanDiffVulnerabilityCollection(SCAScanVulnerabilityCollection):
    type = SCAVulnerabilityCollectionType.DIFF
    result: Optional[SCAScanDiffOutput]

    @property
    def has_results(self) -> bool:
        """
        Checks if at least one package in a location defined in added_vuls
        has one vulnerability
        """
        return self.result is not None and any(
            len(package.vulns) > 0
            for location in self.result.added_vulns
            for package in location.package_vulns
        )



from typing import Any, Dict, cast

from pygitguardian.iac_models import IaCDiffScanEntities, IaCFileResult

from ggshield.verticals.iac.collection.iac_diff_scan_collection import (
    IaCDiffScanCollection,
)
from ggshield.verticals.iac.collection.iac_path_scan_collection import (
    IaCPathScanCollection,
)
from ggshield.verticals.iac.output.iac_output_handler import IaCOutputHandler
from ggshield.verticals.iac.output.schemas import (
    IaCJSONScanDiffResultSchema,
    IaCJSONScanResultSchema,
)


class IaCJSONOutputHandler(IaCOutputHandler):
    def _process_scan_impl(self, scan: IaCPathScanCollection) -> str:
        scan_dict = IaCJSONOutputHandler.create_scan_dict(scan)
        text = IaCJSONScanResultSchema().dumps(scan_dict)
        return cast(str, text)

    def _process_skip_scan_impl(self) -> str:
        return "{}"

    def _process_skip_diff_scan_impl(self) -> str:
        return "{}"

    def _process_diff_scan_impl(self, scan: IaCDiffScanCollection) -> str:
        scan_dict = IaCJSONOutputHandler.create_diff_scan_dict(scan)
        text = IaCJSONScanDiffResultSchema().dumps(scan_dict)
        return cast(str, text)

    @staticmethod
    def create_scan_dict(scan: IaCPathScanCollection) -> Dict[str, Any]:
        if scan.result is None:
            return {
                "id": scan.id,
                "type": scan.type.value,
                "total_incidents": 0,
                "entities_with_incidents": [],
            }
        scan_dict = scan.result.to_dict()
        scan_dict["total_incidents"] = 0

        for entity in scan_dict["entities_with_incidents"]:
            total_incidents = len(entity["incidents"])
            entity["total_incidents"] = total_incidents
            scan_dict["total_incidents"] += total_incidents

        return scan_dict

    @staticmethod
    def create_diff_scan_dict(scan: IaCDiffScanCollection) -> Dict[str, Any]:
        ret: Dict[str, Any] = {
            "added_vulns": [],
            "persisting_vulns": [],
            "removed_vulns": [],
        }

        def file_result_transform(file_result: IaCFileResult) -> Dict[str, Any]:
            return dict(
                filename=file_result.filename,
                incidents=file_result.incidents,
                total_incidents=len(file_result.incidents),
            )

        if scan.result is not None and isinstance(
            scan.result.entities_with_incidents, IaCDiffScanEntities
        ):
            for file_result in scan.result.entities_with_incidents.new:
                ret["added_vulns"].append(file_result_transform(file_result))
            for file_result in scan.result.entities_with_incidents.unchanged:
                ret["persisting_vulns"].append(file_result_transform(file_result))
            for file_result in scan.result.entities_with_incidents.deleted:
                ret["removed_vulns"].append(file_result_transform(file_result))
        return ret



import platform
import uuid
from dataclasses import dataclass
from typing import Dict, Optional, Union

from ggshield import __version__
from ggshield.utils.os import get_os_info

from .scan_mode import ScanMode


@dataclass
class ScanContext:
    scan_mode: Union[ScanMode, str]
    command_path: str
    extra_headers: Optional[Dict[str, str]] = None

    def __post_init__(self) -> None:
        self.command_id = str(uuid.uuid4())
        self.os_name, self.os_version = get_os_info()
        self.python_version = platform.python_version()

    def get_http_headers(self) -> Dict[str, str]:
        """
        Returns the extra headers to send in HTTP requests.
        If `command_id` is not None, a `GGShield-Command-Id` header will be sent.
        Adds the "GGShield-" prefix to the header's names.
        """
        headers = {
            "Version": __version__,
            "Command-Path": self.command_path,
            "Command-Id": self.command_id,
            "OS-Name": self.os_name,
            "OS-Version": self.os_version,
            "Python-Version": self.python_version,
        }
        if self.extra_headers:
            headers = {**headers, **self.extra_headers}

        return {
            **{f"GGShield-{key}": str(value) for key, value in headers.items()},
            "mode": self.scan_mode.value
            if isinstance(self.scan_mode, ScanMode)
            else self.scan_mode,
        }



from collections import Counter
from dataclasses import dataclass, field
from io import StringIO
from typing import ClassVar, Dict, Generator, List, Optional, Union

from pygitguardian.sca_models import (
    SCALocationVulnerability,
    SCAScanAllOutput,
    SCAScanDiffOutput,
)

from ggshield.core.constants import IncidentSeverity, IncidentStatus
from ggshield.core.text_utils import STYLE, file_info, format_text, pluralize
from ggshield.verticals.sca.collection import (
    SCAScanAllVulnerabilityCollection,
    SCAScanDiffVulnerabilityCollection,
)
from ggshield.verticals.sca.output.handler import SCAOutputHandler


@dataclass
class OutputIncidentData:
    package_full_name: str
    version: str
    severity: str
    summary: str
    fixed_version: Optional[str]
    identifier: str
    cve_ids: List[str]

    def sort_key(self) -> IncidentSeverity:
        """IncidentSeverity defines __lt__ to allow sorting"""
        try:
            incident_severity = IncidentSeverity(self.severity)
        except ValueError:
            return IncidentSeverity.UNKNOWN
        return incident_severity


def init_severity_counter() -> Dict[str, int]:
    return {severity.value: 0 for severity in IncidentSeverity}


@dataclass
class OutputLocationData:
    """
    Aggregate SCA incidents data to be printed in one class
    The counters allow to sort according to the number of incidents with each severity
    """

    location: str
    added: List[OutputIncidentData] = field(default_factory=list)
    removed: List[OutputIncidentData] = field(default_factory=list)
    added_counter: Dict[str, int] = field(default_factory=init_severity_counter)
    removed_counter: Dict[str, int] = field(default_factory=init_severity_counter)

    def sort_key(self) -> List[int]:
        return list(self.added_counter.values()) + list(self.removed_counter.values())


def populate_incidents_list(
    result_list: List[SCALocationVulnerability],
    locations_with_incidents: Dict[str, OutputLocationData],
    as_removed_incidents: bool = False,
) -> None:
    """
    Populates dict locations with incidents as OutputLocationData from a list
    of SCALocationVulnerability.
    By default, consider incidents as added incidents.
    If `as_removed_incidents=True`, set them as removed incidents.
    """

    if as_removed_incidents:
        incident_list_key = "removed"
        counter_key = "removed_counter"
    else:
        incident_list_key = "added"
        counter_key = "added_counter"

    for vuln_dep_file in result_list:
        location = vuln_dep_file.location
        if location not in locations_with_incidents:
            locations_with_incidents[location] = OutputLocationData(location=location)

        # Update counter of vulns to sort locations by number of incidents
        for pkg_version_with_vuln in vuln_dep_file.package_vulns:
            for vuln in pkg_version_with_vuln.vulns:
                incidents_list = locations_with_incidents[location].__getattribute__(
                    incident_list_key
                )
                incidents_list.append(
                    OutputIncidentData(
                        package_full_name=pkg_version_with_vuln.package_full_name,
                        version=pkg_version_with_vuln.version,
                        severity=vuln.severity,
                        summary=vuln.summary,
                        fixed_version=vuln.fixed_version,
                        identifier=vuln.identifier,
                        cve_ids=vuln.cve_ids,
                    )
                )

                # Update counter of vulns to sort locations by number of incidents
                locations_with_incidents[location].__getattribute__(counter_key)[
                    vuln.severity.lower()
                ] += 1


def get_sorted_locations(
    result: Union[SCAScanDiffOutput, SCAScanAllOutput], with_removed: bool = False
) -> Generator[OutputLocationData, None, None]:
    """
    Retrieve the locations from a SCAScan output sorted by number of incidents
    for each severity
    Parses removed incidents only if `with_removed` is set to True, in case of a scan diff.
    """

    # Check arguments validity
    if isinstance(result, SCAScanAllOutput) and with_removed:
        raise ValueError(
            "get_sorted_locations cannot be called with with_removed in case of a scan all."
        )

    # First generates a dict of locations with the associated number of incidents
    locations_with_vulns: Dict[str, OutputLocationData] = {}

    # Handle added or existing vulnerabilities
    populate_incidents_list(
        result_list=(
            result.added_vulns
            if isinstance(result, SCAScanDiffOutput)
            else result.found_package_vulns
        ),
        locations_with_incidents=locations_with_vulns,
    )

    # Handle removed ones if required, isinstance check required for pyright
    if isinstance(result, SCAScanDiffOutput) and with_removed:
        populate_incidents_list(
            result_list=result.removed_vulns,
            locations_with_incidents=locations_with_vulns,
            as_removed_incidents=True,
        )

    # Finally yield locations in sorted order
    for vuln_loc in sorted(
        locations_with_vulns.values(), key=lambda elt: elt.sort_key(), reverse=True
    ):
        yield vuln_loc


class SCATextOutputHandler(SCAOutputHandler):
    nb_lines: ClassVar[int] = 3

    def _process_scan_all_impl(self, scan: SCAScanAllVulnerabilityCollection) -> str:
        scan_buf = StringIO()

        # We precise is not None as __bool__ have been overriden for pygitguardian
        # Base class
        if scan.result is not None:
            # List incidents if any
            for file_result in get_sorted_locations(scan.result):
                scan_buf.write(self.file_header(file_result))
                scan_buf.write(
                    self.process_file_result_incidents(file_result.added, prefix="")
                )

            # Show no incidents if none
            if len(scan.result.found_package_vulns) == 0:
                scan_buf.write(no_sca_vulnerabilities())
        return scan_buf.getvalue()

    def _process_scan_diff_impl(self, scan: SCAScanDiffVulnerabilityCollection) -> str:
        """
        In the non verbose case, only new vulnerabilities are shown in
        text output.
        """
        scan_buf = StringIO()

        # We precise is not None as __bool__ have been overriden for pygitguardian
        # Base class
        if scan.result is not None:
            # List added incidents if any
            for file_result in get_sorted_locations(
                scan.result, with_removed=self.verbose
            ):
                scan_buf.write(self.file_header(file_result))
                # Show added vulnerabilities
                scan_buf.write(
                    self.process_file_result_incidents(file_result.added, prefix="NEW")
                )

                # Skip the removed vulns if not verbose
                if not self.verbose:
                    continue

                scan_buf.write(
                    self.process_file_result_incidents(
                        file_result.removed,
                        counter_init=len(file_result.added) + 1,
                        prefix="REMOVED",
                        detailed=False,
                    )
                )

            # Show no incidents if none
            if len(scan.result.added_vulns) == 0:
                scan_buf.write(no_sca_vulnerability_added())

            scan_buf.write(diff_scan_summary(scan.result))

        return scan_buf.getvalue()

    def file_header(self, file_result: OutputLocationData) -> str:
        result_buf = StringIO()
        result_buf.write(
            file_info(
                file_result.location,
                # Compute number of added incidents in the file
                sum(nbr for nbr in file_result.added_counter.values()),
                incident_status=IncidentStatus.DETECTED,
            )
        )
        return result_buf.getvalue()

    def process_file_result_incidents(
        self,
        incident_list: List[OutputIncidentData],
        counter_init: int = 1,
        prefix: str = "",
        detailed: bool = True,
    ) -> str:
        result_buf = StringIO()

        incident_n = counter_init
        for incident in sorted(incident_list, key=lambda x: x.sort_key()):
            result_buf.write(sca_incident_header(incident_n, incident, prefix=prefix))

            result_buf.write(sca_incident_severity_line(incident))
            if detailed:
                result_buf.write(sca_incident_summary_line(incident))
                result_buf.write(sca_incident_fix_version_line(incident))
            result_buf.write(sca_incident_identifier(incident))
            result_buf.write(sca_incident_cve_ids(incident))

            incident_n += 1

        return result_buf.getvalue()


def no_sca_vulnerabilities() -> str:
    """
    Build a message if no SCA vulnerabilities were found.
    """
    return format_text("\nNo SCA vulnerability has been found.\n", STYLE["no_secret"])


def no_sca_vulnerability_added() -> str:
    """
    Build a message if no SCA vulnerabilities were added.
    """
    return format_text("\nNo SCA vulnerability has been added.\n", STYLE["no_secret"])


def sca_incident_header(
    incident_n: int,
    package_version: OutputIncidentData,
    prefix: Optional[str] = None,
) -> str:
    return "\n{}{} Incident {} ({}): {}@{}\n".format(
        format_text(">>>", STYLE["detector_line_start"]),
        "" if prefix is None else f" {prefix}:",
        incident_n,
        format_text("SCA", STYLE["detector"]),
        format_text(package_version.package_full_name, STYLE["detector"]),
        format_text(package_version.version, STYLE["detector"]),
    )


def sca_incident_severity_line(vulnerability: OutputIncidentData) -> str:
    """
    Returns the severity line, with associated style
    """
    if vulnerability.severity.lower() == IncidentSeverity.CRITICAL:
        severity_string = IncidentSeverity.CRITICAL.value.capitalize()
        style = STYLE["sca_vulnerability_critical"]
    elif vulnerability.severity.lower() == IncidentSeverity.HIGH:
        severity_string = IncidentSeverity.HIGH.value.capitalize()
        style = STYLE["sca_vulnerability_high"]
    elif vulnerability.severity.lower() == IncidentSeverity.MEDIUM:
        severity_string = IncidentSeverity.MEDIUM.value.capitalize()
        style = STYLE["sca_vulnerability_medium"]
    elif vulnerability.severity.lower() == IncidentSeverity.LOW:
        severity_string = IncidentSeverity.LOW.value.capitalize()
        style = STYLE["sca_vulnerability_low"]

    else:  # In the other case, print `severity` directly
        severity_string = vulnerability.severity
        style = STYLE["sca_vulnerability_unknown"]

    return "Severity: {}\n".format(
        format_text(severity_string, style),
    )


def sca_incident_summary_line(vulnerability: OutputIncidentData) -> str:
    return f"Summary: {vulnerability.summary}\n"


def sca_incident_fix_version_line(vulnerability: OutputIncidentData) -> str:
    """
    Return the fixed version information if exists.
    """
    if vulnerability.fixed_version is not None:
        return f"A fix is available at version {vulnerability.fixed_version}\n"

    return "No fix is currently available.\n"


def sca_incident_cve_ids(vulnerability: OutputIncidentData) -> str:
    if not vulnerability.cve_ids:
        cve_str = "-"
    else:
        cve_str = ", ".join(vulnerability.cve_ids)
    return f"CVE IDs: {cve_str}\n"


def sca_incident_identifier(vulnerability: OutputIncidentData) -> str:
    identifier = vulnerability.identifier
    # TODO Remove clause when identifier field is in production on backend side (v2.36)
    if identifier is None:
        return ""
    return f"Identifier: {identifier}\n"


def diff_scan_summary(result: SCAScanDiffOutput) -> str:
    added_counter = Counter(
        vuln.severity
        for location in result.added_vulns
        for pkg in location.package_vulns
        for vuln in pkg.vulns
    )
    num_added = sum(added_counter.values())

    removed_counter = Counter(
        vuln.severity
        for location in result.removed_vulns
        for pkg in location.package_vulns
        for vuln in pkg.vulns
    )
    num_removed = sum(removed_counter.values())

    def label_incident(n: int) -> str:
        return pluralize("incident", n, "incidents")

    result_buf = StringIO()

    result_buf.write("\nSummary of changes:\n")
    # removed vulns
    result_buf.write(f"[-] {num_removed} {label_incident(num_removed)} deleted")
    if num_removed:
        severity_counters = ", ".join(
            f"{severity.value.capitalize()}: {removed_counter[severity.value]}"
            for severity in IncidentSeverity
            if severity.value in removed_counter
        )
        result_buf.write(f" ({severity_counters})")
    result_buf.write("\n")

    # added vulns
    result_buf.write(f"[+] {num_added} {label_incident(num_added)} detected")
    if num_added:
        severity_counters = ", ".join(
            f"{severity.value.capitalize()}: {added_counter[severity.value]}"
            for severity in IncidentSeverity
            if severity.value in added_counter
        )
        result_buf.write(f" ({severity_counters})")
    result_buf.write("\n")

    return result_buf.getvalue()



from pathlib import Path
from typing import Any, Optional, Sequence

import click

from ggshield.cmd.iac.scan.all import display_iac_scan_all_result, iac_scan_all
from ggshield.cmd.iac.scan.diff import display_iac_scan_diff_result, iac_scan_diff
from ggshield.cmd.iac.scan.iac_scan_common_options import (
    add_iac_scan_common_options,
    update_context,
)
from ggshield.cmd.utils.common_decorators import display_beta_warning, exception_wrapper
from ggshield.cmd.utils.common_options import all_option, directory_argument
from ggshield.core.config import Config
from ggshield.core.git_hooks.ci import get_current_and_previous_state_from_ci_env
from ggshield.core.git_hooks.ci.supported_ci import SupportedCI
from ggshield.core.scan.scan_mode import ScanMode


@click.command()
@add_iac_scan_common_options()
@all_option
@directory_argument
@click.pass_context
@display_beta_warning
@exception_wrapper
def scan_ci_cmd(
    ctx: click.Context,
    exit_zero: bool,
    minimum_severity: str,
    ignore_policies: Sequence[str],
    ignore_paths: Sequence[str],
    scan_all: bool,
    directory: Optional[Path] = None,
    **kwargs: Any,
) -> int:
    """
    Scan in CI for IaC vulnerabilities. By default, it will return vulnerabilities added in the new commits.

    The scan is successful if no *new* IaC vulnerability was found, unless `--all` is used,
    in which case the scan is only successful if no IaC vulnerability (old and new) was found.
    """
    config: Config = ctx.obj["config"]
    if directory is None:
        directory = Path().resolve()
    update_context(ctx, exit_zero, minimum_severity, ignore_policies, ignore_paths)
    ci_mode = SupportedCI.from_ci_env()

    if scan_all:
        result = iac_scan_all(
            ctx, directory, scan_mode=ScanMode.CI_ALL, ci_mode=ci_mode
        )
        return display_iac_scan_all_result(ctx, directory, result)

    current_commit, previous_commit = get_current_and_previous_state_from_ci_env(
        config.user_config.verbose
    )

    result = iac_scan_diff(
        ctx,
        directory,
        previous_commit,
        current_ref=current_commit,
        include_staged=True,
        scan_mode=ScanMode.CI_DIFF,
        ci_mode=ci_mode,
    )
    return display_iac_scan_diff_result(ctx, directory, result)



from typing import Any

import click

from ggshield.cmd.utils.common_options import add_common_options

from .scan import scan_group


@click.group(commands={"scan": scan_group})
@add_common_options()
def sca_group(**kwargs: Any) -> None:
    """Commands to work with SCA."""



import os
from typing import List, Tuple

import click

from ggshield.core.errors import UnexpectedError
from ggshield.utils.git_shell import EMPTY_SHA, get_list_commit_SHA

from .previous_commit import (
    github_pull_request_previous_commit_sha,
    github_push_previous_commit_sha,
    gitlab_push_previous_commit_sha,
)
from .supported_ci import SupportedCI


def collect_commit_range_from_ci_env(
    verbose: bool,
) -> Tuple[List[str], SupportedCI]:
    supported_ci = SupportedCI.from_ci_env()
    try:
        fcn = COLLECT_COMMIT_RANGE_FUNCTIONS[supported_ci]
    except KeyError:
        raise UnexpectedError(f"Not implemented for {supported_ci.value}")

    return fcn(verbose), supported_ci


def jenkins_range(verbose: bool) -> List[str]:  # pragma: no cover
    head_commit = os.getenv("GIT_COMMIT")
    previous_commit = os.getenv("GIT_PREVIOUS_COMMIT")

    if verbose:
        click.echo(
            f"\tGIT_COMMIT: {head_commit}" f"\nGIT_PREVIOUS_COMMIT: {previous_commit}",
            err=True,
        )

    if previous_commit:
        commit_list = get_list_commit_SHA(f"{previous_commit}...{head_commit}")
        if commit_list:
            return commit_list

    commit_list = get_list_commit_SHA(f"{head_commit}~1...")
    if commit_list:
        return commit_list

    raise UnexpectedError(
        "Unable to get commit range. Please submit an issue with the following info:\n"
        "\tRepository URL: <Fill if public>\n"
        f"\tGIT_COMMIT: {head_commit}"
        f"\tGIT_PREVIOUS_COMMIT: {previous_commit}"
    )


def travis_range(verbose: bool) -> List[str]:  # pragma: no cover
    commit_range = os.getenv("TRAVIS_COMMIT_RANGE")
    commit_sha = os.getenv("TRAVIS_COMMIT", "HEAD")

    if verbose:
        click.echo(
            f"TRAVIS_COMMIT_RANGE: {commit_range}" f"\nTRAVIS_COMMIT: {commit_sha}",
            err=True,
        )

    if commit_range:
        commit_list = get_list_commit_SHA(commit_range)
        if commit_list:
            return commit_list

    commit_list = get_list_commit_SHA(f"{commit_sha}~1...")
    if commit_list:
        return commit_list

    raise UnexpectedError(
        "Unable to get commit range. Please submit an issue with the following info:\n"
        "\tRepository URL: <Fill if public>\n"
        f"\tTRAVIS_COMMIT_RANGE: {commit_range}"
        f"\tTRAVIS_COMMIT: {commit_sha}"
    )


def bitbucket_pipelines_range(verbose: bool) -> List[str]:  # pragma: no cover
    commit_sha = os.getenv("BITBUCKET_COMMIT", "HEAD")
    if verbose:
        click.echo(f"BITBUCKET_COMMIT: {commit_sha}", err=True)

    commit_list = get_list_commit_SHA(f"{commit_sha}~1...")
    if commit_list:
        return commit_list

    raise UnexpectedError(
        "Unable to get commit range. Please submit an issue with the following info:\n"
        "  Repository URL: <Fill if public>\n"
        f"  CI_COMMIT_SHA: {commit_sha}"
    )


def circle_ci_range(verbose: bool) -> List[str]:  # pragma: no cover
    """
    # Extract commit range (or single commit)
    COMMIT_RANGE=$(echo "${CIRCLE_COMPARE_URL}" | cut -d/ -f7)

    # Fix single commit, unfortunately we don't always get a commit range from Circle CI
    if [[ $COMMIT_RANGE != *"..."* ]]; then
    COMMIT_RANGE="${COMMIT_RANGE}...${COMMIT_RANGE}"
    fi
    """
    compare_range = os.getenv("CIRCLE_RANGE")
    commit_sha = os.getenv("CIRCLE_SHA1", "HEAD")

    if verbose:
        click.echo(
            f"CIRCLE_RANGE: {compare_range}\nCIRCLE_SHA1: {commit_sha}", err=True
        )

    if compare_range and not compare_range.startswith("..."):
        commit_list = get_list_commit_SHA(compare_range)
        if commit_list:
            return commit_list

    commit_list = get_list_commit_SHA(f"{commit_sha}~1...")
    if commit_list:
        return commit_list

    raise UnexpectedError(
        "Unable to get commit range. Please submit an issue with the following info:\n"
        "\tRepository URL: <Fill if public>\n"
        f"\tCIRCLE_RANGE: {compare_range}\n"
        f"\tCIRCLE_SHA1: {commit_sha}"
    )


def gitlab_ci_range(verbose: bool) -> List[str]:  # pragma: no cover
    before_sha = gitlab_push_previous_commit_sha()
    commit_sha = os.getenv("CI_COMMIT_SHA", "HEAD")
    merge_request_target_branch = os.getenv("CI_MERGE_REQUEST_TARGET_BRANCH_NAME")

    if verbose:
        click.echo(
            f"CI_MERGE_REQUEST_TARGET_BRANCH_NAME: {merge_request_target_branch}\n"
            f"CI_COMMIT_BEFORE_SHA: {before_sha}\n"
            f"CI_COMMIT_SHA: {commit_sha}",
            err=True,
        )

    if before_sha and before_sha != EMPTY_SHA:
        commit_list = get_list_commit_SHA(f"{before_sha}~1...")
        if commit_list:
            return commit_list

    if merge_request_target_branch and merge_request_target_branch != EMPTY_SHA:
        commit_list = get_list_commit_SHA(f"origin/{merge_request_target_branch}...")
        if commit_list:
            return commit_list

    commit_list = get_list_commit_SHA(f"{commit_sha}~1...")
    if commit_list:
        return commit_list

    raise UnexpectedError(
        "Unable to get commit range. Please submit an issue with the following info:\n"
        "  Repository URL: <Fill if public>\n"
        f"  CI_MERGE_REQUEST_TARGET_BRANCH_NAME: {merge_request_target_branch}\n"
        f"  CI_COMMIT_BEFORE_SHA: {before_sha}\n"
        f"  CI_COMMIT_SHA: {commit_sha}"
    )


def github_actions_range(verbose: bool) -> List[str]:  # pragma: no cover
    push_before_sha = github_push_previous_commit_sha()
    push_base_sha = os.getenv("GITHUB_PUSH_BASE_SHA")
    pull_req_base_sha = github_pull_request_previous_commit_sha()
    default_branch = os.getenv("GITHUB_DEFAULT_BRANCH")
    head_sha = os.getenv("GITHUB_SHA", "HEAD")

    if verbose:
        click.echo(
            f"github_push_before_sha: {push_before_sha}\n"
            f"github_push_base_sha: {push_base_sha}\n"
            f"github_pull_base_sha: {pull_req_base_sha}\n"
            f"github_default_branch: {default_branch}\n"
            f"github_head_sha: {head_sha}",
            err=True,
        )

    # The PR base sha has to be checked before the push_before_sha
    # because the first one is only populated in case of PR
    # whereas push_before_sha can be populated in PR in case of
    # push force event in a PR
    if pull_req_base_sha and pull_req_base_sha != EMPTY_SHA:
        commit_list = get_list_commit_SHA(f"{pull_req_base_sha}..")
        if commit_list:
            return commit_list

    if push_before_sha and push_before_sha != EMPTY_SHA:
        commit_list = get_list_commit_SHA(f"{push_before_sha}...")
        if commit_list:
            return commit_list

    if push_base_sha and push_base_sha != "null":
        commit_list = get_list_commit_SHA(f"{push_base_sha}...")
        if commit_list:
            return commit_list

    if default_branch:
        commit_list = get_list_commit_SHA(f"{default_branch}...")
        if commit_list:
            return commit_list

    if head_sha:
        commit_list = get_list_commit_SHA(f"{head_sha}~1...")
        if commit_list:
            return commit_list

    raise UnexpectedError(
        "Unable to get commit range. Please submit an issue with the following info:\n"
        "  Repository URL: <Fill if public>\n"
        f"github_push_before_sha: {push_before_sha}\n"
        f"github_push_base_sha: {push_base_sha}\n"
        f"github_pull_base_sha: {pull_req_base_sha}\n"
        f"github_default_branch: {default_branch}\n"
        f"github_head_sha: {head_sha}"
    )


def drone_range(verbose: bool) -> List[str]:  # pragma: no cover
    before_sha = os.getenv("DRONE_COMMIT_BEFORE")

    if verbose:
        click.echo(f"DRONE_COMMIT_BEFORE: {before_sha}\n", err=True)

    if before_sha and before_sha != EMPTY_SHA:
        commit_list = get_list_commit_SHA(f"{before_sha}..")
        if commit_list:
            return commit_list

    raise UnexpectedError(
        "Unable to get commit range. Please submit an issue with the following info:\n"
        "  Repository URL: <Fill if public>\n"
        f"  DRONE_COMMIT_BEFORE: {before_sha}"
    )


def azure_range(verbose: bool) -> List[str]:  # pragma: no cover
    head_commit = os.getenv("BUILD_SOURCEVERSION")

    if verbose:
        click.echo(f"BUILD_SOURCEVERSION: {head_commit}\n", err=True)

    if head_commit:
        commit_list = get_list_commit_SHA(f"{head_commit}~1...")
        if commit_list:
            return commit_list

    raise UnexpectedError(
        "Unable to get commit range. Please submit an issue with the following info:\n"
        "  Repository URL: <Fill if public>\n"
        f"  BUILD_SOURCEVERSION: {head_commit}"
    )


COLLECT_COMMIT_RANGE_FUNCTIONS = {
    SupportedCI.AZURE: azure_range,
    SupportedCI.DRONE: drone_range,
    SupportedCI.GITHUB: github_actions_range,
    SupportedCI.GITLAB: gitlab_ci_range,
    SupportedCI.CIRCLECI: circle_ci_range,
    SupportedCI.BITBUCKET: bitbucket_pipelines_range,
    SupportedCI.TRAVIS: travis_range,
    SupportedCI.JENKINS: jenkins_range,
}



from typing import Any

import click

from ggshield.cmd.iac.scan.all import scan_all_cmd
from ggshield.cmd.iac.scan.ci import scan_ci_cmd
from ggshield.cmd.iac.scan.diff import scan_diff_cmd
from ggshield.cmd.iac.scan.iac_scan_common_options import add_iac_scan_common_options
from ggshield.cmd.iac.scan.precommit import scan_pre_commit_cmd
from ggshield.cmd.iac.scan.prepush import scan_pre_push_cmd
from ggshield.cmd.iac.scan.prereceive import scan_pre_receive_cmd
from ggshield.cmd.utils.common_options import directory_argument
from ggshield.core.client import create_client_from_config
from ggshield.core.text_utils import display_warning
from ggshield.utils.click import DefaultCommandGroup


@click.group(
    cls=DefaultCommandGroup,
    commands={
        "all": scan_all_cmd,
        "ci": scan_ci_cmd,
        "diff": scan_diff_cmd,
        "pre-commit": scan_pre_commit_cmd,
        "pre-push": scan_pre_push_cmd,
        "pre-receive": scan_pre_receive_cmd,
    },
    invoke_without_command=True,
)
@add_iac_scan_common_options()
@click.pass_context
def iac_scan_group(
    ctx: click.Context,
    **kwargs: Any,
) -> int:
    """Commands to scan various contents."""
    return scan_group_impl(ctx)


# Alias to "ggshield iac scan all"
# Kept for compatibility
# Changes to arguments must be propagated to scan_all_cmd
@iac_scan_group.command(default_command=True)
@add_iac_scan_common_options()
@directory_argument
@click.pass_context
def default_command(ctx: click.Context, **kwargs: Any) -> int:
    """Deprecated. Use `ggshield iac scan all` instead"""
    display_warning("Deprecated. Please use 'ggshield iac scan all' instead")
    result: int = scan_all_cmd.invoke(ctx)
    return result


def scan_group_impl(ctx: click.Context) -> int:
    """Implementation for scan_group(). Must be a separate function so that its code can
    be reused from the deprecated `cmd.scan` package."""
    if ctx.invoked_subcommand is None:
        scan_all_cmd.invoke(ctx)

    ctx.obj["client"] = create_client_from_config(ctx.obj["config"])

    return 0



from typing import Any

import click

from ggshield.cmd.sca.scan.all import scan_all_cmd
from ggshield.cmd.sca.scan.ci import scan_ci_cmd
from ggshield.cmd.sca.scan.diff import scan_diff_cmd
from ggshield.cmd.sca.scan.precommit import scan_pre_commit_cmd
from ggshield.cmd.sca.scan.prepush import scan_pre_push_cmd
from ggshield.cmd.sca.scan.prereceive import scan_pre_receive_cmd


@click.group(
    commands={
        "pre-commit": scan_pre_commit_cmd,
        "all": scan_all_cmd,
        "ci": scan_ci_cmd,
        "pre-push": scan_pre_push_cmd,
        "diff": scan_diff_cmd,
        "pre-receive": scan_pre_receive_cmd,
    }
)
@click.pass_context
def scan_group(*args, **kwargs: Any) -> None:
    """Perform a SCA scan."""



from typing import Optional

from pygitguardian.iac_models import IaCDiffScanResult

from ggshield.verticals.iac.collection.iac_scan_collection import (
    CollectionType,
    IaCScanCollection,
)


class IaCDiffScanCollection(IaCScanCollection):
    type = CollectionType.DiffScan
    result: Optional[IaCDiffScanResult]

    @property
    def has_results(self) -> bool:
        return self.result is not None and bool(self.result.entities_with_incidents.new)



from pathlib import Path
from typing import Any, List

import click

from ggshield.cmd.secret.scan.secret_scan_common_options import (
    add_secret_scan_common_options,
    create_output_handler,
)
from ggshield.cmd.utils.common_decorators import exception_wrapper
from ggshield.cmd.utils.files import check_directory_not_ignored
from ggshield.core.config import Config
from ggshield.core.scan import ScanContext, ScanMode
from ggshield.core.scan.file import get_files_from_paths
from ggshield.utils.click import RealPath
from ggshield.verticals.secret import (
    RichSecretScannerUI,
    SecretScanCollection,
    SecretScanner,
)


@click.command()
@click.argument(
    "paths", nargs=-1, type=RealPath(exists=True, resolve_path=True), required=True
)
@click.option("--recursive", "-r", is_flag=True, help="Scan directory recursively.")
@click.option("--yes", "-y", is_flag=True, help="Confirm recursive scan.")
@add_secret_scan_common_options()
@click.pass_context
@exception_wrapper
def path_cmd(
    ctx: click.Context,
    paths: List[Path],
    recursive: bool,
    yes: bool,
    **kwargs: Any,
) -> int:  # pragma: no cover
    """
    Scan files and directories.
    """
    config: Config = ctx.obj["config"]
    output_handler = create_output_handler(ctx)
    verbose = config.user_config.verbose

    for path in paths:
        check_directory_not_ignored(path, ctx.obj["exclusion_regexes"])

    files = get_files_from_paths(
        paths=paths,
        exclusion_regexes=ctx.obj["exclusion_regexes"],
        recursive=recursive,
        yes=yes,
        display_scanned_files=verbose,
        display_binary_files=verbose,
        # when scanning a path explicitly we should not care if it is a git repository or not
        ignore_git=True,
    )

    with RichSecretScannerUI(len(files), dataset_type="Path") as ui:
        scan_context = ScanContext(
            scan_mode=ScanMode.PATH,
            command_path=ctx.command_path,
        )

        scanner = SecretScanner(
            client=ctx.obj["client"],
            cache=ctx.obj["cache"],
            ignored_matches=config.user_config.secret.ignored_matches,
            scan_context=scan_context,
            ignored_detectors=config.user_config.secret.ignored_detectors,
        )
        results = scanner.scan(files, scanner_ui=ui)
    scan = SecretScanCollection(
        id=" ".join(str(x) for x in paths), type="path_scan", results=results
    )

    return output_handler.process_scan(scan)



from dataclasses import dataclass
from typing import cast

import marshmallow_dataclass
from pygitguardian.models import BaseSchema
from pygitguardian.sca_models import SCAScanAllOutput, SCAScanDiffOutput


@dataclass
class SCAJSONScanAllOutput(SCAScanAllOutput):
    total_vulns: int = 0


SCAJSONScanAllOutputSchema = cast(
    BaseSchema,
    marshmallow_dataclass.class_schema(SCAJSONScanAllOutput, base_schema=BaseSchema)(),
)


@dataclass
class SCAJSONScanDiffOutput(SCAScanDiffOutput):
    total_vulns: int = 0


SCAJSONScanDiffOutputSchema = cast(
    BaseSchema,
    marshmallow_dataclass.class_schema(SCAJSONScanDiffOutput, base_schema=BaseSchema)(),
)



from abc import ABC, abstractproperty
from enum import Enum
from typing import Optional, Union

from pygitguardian.iac_models import IaCDiffScanResult, IaCScanResult


IaCResult = Union[IaCScanResult, IaCDiffScanResult]


class CollectionType(Enum):
    Unknown = "unknown"
    PathScan = "path_scan"
    DiffScan = "diff_scan"


class IaCScanCollection(ABC):
    type = CollectionType.Unknown
    id: str
    # Can be None if the scan failed
    result: Optional[IaCResult]

    def __init__(
        self,
        id: str,
        result: Optional[IaCResult],
    ):
        self.id = id
        self.result = result

    @abstractproperty
    def has_results(self) -> bool:
        """
        Whether the scan found problems
        """
        return self.result is not None



from pathlib import Path
from typing import Any, Optional, Sequence

import click

from ggshield.cmd.iac.scan.all import display_iac_scan_all_result, iac_scan_all
from ggshield.cmd.iac.scan.diff import display_iac_scan_diff_result, iac_scan_diff
from ggshield.cmd.iac.scan.iac_scan_common_options import (
    add_iac_scan_common_options,
    update_context,
)
from ggshield.cmd.utils.common_decorators import display_beta_warning, exception_wrapper
from ggshield.cmd.utils.common_options import all_option, directory_argument
from ggshield.core.scan.scan_mode import ScanMode


@click.command()
@add_iac_scan_common_options()
@all_option
@directory_argument
@click.pass_context
@display_beta_warning
@exception_wrapper
def scan_pre_commit_cmd(
    ctx: click.Context,
    exit_zero: bool,
    minimum_severity: str,
    ignore_policies: Sequence[str],
    ignore_paths: Sequence[str],
    scan_all: bool,
    directory: Optional[Path] = None,
    **kwargs: Any,
) -> int:
    """
    Scan a Git repository for changes in IaC vulnerabilities between HEAD and current staged changes.

    The scan is successful if no *new* IaC vulnerability was found, unless `--all` is used,
    in which case the scan is only successful if no IaC vulnerability (old and new) was found.

    By default, the output will show:
    - The number of known IaC vulnerabilities resolved by the changes
    - The number of known IaC vulnerabilities left untouched
    - The number and the list of new IaC vulnerabilities introduced by the changes
    """
    if directory is None:
        directory = Path().resolve()
    update_context(ctx, exit_zero, minimum_severity, ignore_policies, ignore_paths)
    if scan_all:
        result = iac_scan_all(ctx, directory, scan_mode=ScanMode.PRE_COMMIT_ALL)
        return display_iac_scan_all_result(ctx, directory, result)
    result = iac_scan_diff(
        ctx, directory, "HEAD", include_staged=True, scan_mode=ScanMode.PRE_COMMIT_DIFF
    )
    return display_iac_scan_diff_result(ctx, directory, result)



import logging
import sys
from typing import Optional

import pygitguardian


LOG_FORMAT = "%(asctime)s %(levelname)s %(process)x:%(thread)x %(name)s:%(funcName)s:%(lineno)d %(message)s"

logger = logging.getLogger(__name__)


def disable_logs() -> None:
    """By default, disable all logs, because when an error occurs we log an error
    message and also print a human-friendly message using display_errors(). If we don't
    disable all logs, then error logs are printed, resulting in the error being shown
    twice.
    """
    logging.disable()


def setup_debug_logs(*, filename: Optional[str]) -> None:
    """Configure Python logger to log to stderr if filename is None, or to filename if
    it's set.
    """
    # Re-enable logging, reverting the call to disable_logs()
    logging.disable(logging.NOTSET)

    logging.basicConfig(
        filename=filename, level=logging.DEBUG, format=LOG_FORMAT, force=True
    )

    # Silence charset_normalizer, its debug output does not bring much
    logging.getLogger("charset_normalizer").setLevel(logging.WARNING)

    logger.debug("args=%s", sys.argv)
    logger.debug("py-gitguardian=%s", pygitguardian.__version__)



from functools import wraps

import click

from ggshield.core.config import Config
from ggshield.core.errors import handle_exception
from ggshield.core.text_utils import display_warning


def exception_wrapper(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception as error:
            ctx = next(arg for arg in args if isinstance(arg, click.Context))
            config: Config = ctx.obj["config"]
            return handle_exception(error, config.user_config.verbose)

    return wrapper


def display_beta_warning(func):
    """
    Displays warning about new verticals' commands being in beta.
    """

    @wraps(func)
    def func_with_beta_warning(*args, **kwargs):
        display_warning(
            "This feature is still in beta, its behavior may change in future versions."
        )
        return func(*args, **kwargs)

    return func_with_beta_warning



from .collection import (
    SCAScanAllVulnerabilityCollection,
    SCAScanDiffVulnerabilityCollection,
    SCAScanVulnerabilityCollection,
    SCAVulnerabilityCollectionType,
)


__all__ = [
    "SCAVulnerabilityCollectionType",
    "SCAScanVulnerabilityCollection",
    "SCAScanAllVulnerabilityCollection",
    "SCAScanDiffVulnerabilityCollection",
]



import re
from pathlib import Path
from typing import Set

from click import UsageError

from ggshield.utils.files import is_filepath_excluded


def check_directory_not_ignored(directory: Path, exclusion_regexes: Set[re.Pattern]):
    if is_filepath_excluded(directory, exclusion_regexes):
        raise UsageError("An ignored file or directory cannot be scanned.")



from dataclasses import dataclass
from pathlib import Path
from re import Pattern
from typing import Iterable, Optional, Set, Type

import click
from pygitguardian import GGClient
from pygitguardian.models import Detail

from ggshield.cmd.utils.common_options import use_json
from ggshield.core.config import Config
from ggshield.core.errors import APIKeyCheckError
from ggshield.core.tar_utils import INDEX_REF, tar_from_ref_and_filepaths
from ggshield.core.text_utils import display_error
from ggshield.utils.files import is_filepath_excluded
from ggshield.utils.git_shell import get_filepaths_from_ref, get_staged_filepaths
from ggshield.verticals.iac.filter import is_iac_file_path
from ggshield.verticals.iac.output import (
    IaCJSONOutputHandler,
    IaCOutputHandler,
    IaCTextOutputHandler,
)


@dataclass
class IaCSkipScanResult:
    id: str = ""


def create_output_handler(ctx: click.Context) -> IaCOutputHandler:
    """Read objects defined in ctx.obj and create the appropriate OutputHandler
    instance"""
    output_handler_cls: Type[IaCOutputHandler]
    if use_json(ctx):
        output_handler_cls = IaCJSONOutputHandler
    else:
        output_handler_cls = IaCTextOutputHandler
    config: Config = ctx.obj["config"]
    return output_handler_cls(verbose=config.user_config.verbose)


def handle_scan_error(client: GGClient, detail: Detail) -> None:
    if detail.status_code == 401:
        raise APIKeyCheckError(client.base_uri, "Invalid API key.")
    display_error("\nError scanning.")
    display_error(str(detail))


def get_git_filepaths(directory: Path, ref: str) -> Iterable[Path]:
    return (
        get_staged_filepaths(str(directory))
        if ref == INDEX_REF
        else get_filepaths_from_ref(ref, str(directory))
    )


def _accept_iac_file_on_path(
    path: Path, directory: Path, exclusion_regexes: Optional[Set[Pattern]] = None
) -> bool:
    return is_iac_file_path(path) and (
        exclusion_regexes is None
        or not is_filepath_excluded(directory / path, exclusion_regexes)
    )


def filter_iac_filepaths(
    directory: Path,
    filepaths: Iterable[Path],
    exclusion_regexes: Optional[Set[Pattern]] = None,
) -> Iterable[Path]:
    # You can filter based on file's content here
    # using read_git_file (result will be cached)
    # You should filter on path first, in order to read
    # the content only if necessary
    return [
        path
        for path in filepaths
        if _accept_iac_file_on_path(
            path, directory, exclusion_regexes=exclusion_regexes
        )
    ]


def get_iac_tar(directory: Path, ref: str, exclusion_regexes: Set[Pattern]) -> bytes:
    filepaths = get_git_filepaths(directory, ref)
    filtered_paths = filter_iac_filepaths(
        directory=directory,
        filepaths=filepaths,
        exclusion_regexes=exclusion_regexes,
    )

    return tar_from_ref_and_filepaths(ref, filtered_paths, wd=str(directory))



import os
from typing import Optional

import click

from ggshield.core.errors import UnexpectedError
from ggshield.utils.git_shell import (
    EMPTY_SHA,
    get_last_commit_sha_of_branch,
    get_list_commit_SHA,
    git,
)

from .supported_ci import SupportedCI


def get_previous_commit_from_ci_env(
    verbose: bool,
) -> Optional[str]:
    """
    Returns the previous HEAD sha of the targeted branch.
    Returns None if there was no commit before.
    """
    supported_ci = SupportedCI.from_ci_env()
    try:
        fcn = PREVIOUS_COMMIT_SHA_FUNCTIONS[supported_ci]
    except KeyError:
        raise UnexpectedError(f"Not implemented for {supported_ci.value}")

    return fcn(verbose)


def github_previous_commit_sha(verbose: bool) -> Optional[str]:
    push_before_sha = github_push_previous_commit_sha()
    pull_req_base_sha = github_pull_request_previous_commit_sha()
    head_sha = os.getenv("GITHUB_SHA", "HEAD")
    event_name = os.getenv("GITHUB_EVENT_NAME")

    if verbose:
        click.echo(
            f"github_push_before_sha: {push_before_sha}\n"
            f"github_pull_base_sha: {pull_req_base_sha}\n",
            err=True,
        )

    # The PR base sha has to be checked before the push_before_sha
    # because the first one is only populated in case of PR
    # whereas push_before_sha can be populated in PR in case of
    # push force event in a PR
    if pull_req_base_sha:
        return pull_req_base_sha

    if push_before_sha and push_before_sha != EMPTY_SHA:
        return push_before_sha

    if head_sha and event_name == "push":
        if verbose:
            click.echo("Could not find previous commit for current branch.")
            click.echo("Current branch may have been just pushed.")
            click.echo("Only scan last commit.")
        last_commits = get_list_commit_SHA(f"{head_sha}~1", max_count=1)
        if len(last_commits) == 1:
            return last_commits[0]

    raise UnexpectedError(
        "Unable to get previous commit. Please submit an issue with the following info:\n"
        "  Repository URL: <Fill if public>\n"
        f"github_push_before_sha: {push_before_sha}\n"
        f"github_pull_base_sha: {pull_req_base_sha}\n"
    )


def github_push_previous_commit_sha() -> Optional[str]:
    push_before_sha = os.getenv("GITHUB_PUSH_BEFORE_SHA")

    if not push_before_sha:
        return None

    return push_before_sha


def github_pull_request_previous_commit_sha() -> Optional[str]:
    targeted_branch = os.getenv("GITHUB_BASE_REF")

    # Not in a pull request workflow
    if targeted_branch is None:
        return None

    return get_last_commit_sha_of_branch(f"remotes/origin/{targeted_branch}")


def gitlab_previous_commit_sha(verbose: bool) -> Optional[str]:
    push_before_sha = gitlab_push_previous_commit_sha()
    merge_req_base_sha = gitlab_merge_request_previous_commit_sha(verbose)

    if verbose:
        click.echo(
            f"gitlab_push_before_sha: {push_before_sha}\n"
            f"gitlab_merge_base_sha: {merge_req_base_sha}\n",
            err=True,
        )

    # push_before_sha is always EMPTY_SHA in MR pipeline according with
    # https://docs.gitlab.com/ee/ci/variables/predefined_variables.html
    if push_before_sha == EMPTY_SHA and merge_req_base_sha:
        # Targeted branch is empty
        if merge_req_base_sha == EMPTY_SHA:
            return None
        return merge_req_base_sha

    if push_before_sha and push_before_sha != EMPTY_SHA:
        return push_before_sha

    raise UnexpectedError(
        "Unable to get previous commit. Please submit an issue with the following info:\n"
        "  Repository URL: <Fill if public>\n"
        f"gitlab_push_before_sha: {push_before_sha}\n"
        f"gitlab_pull_base_sha: {merge_req_base_sha}\n"
    )


def gitlab_push_previous_commit_sha() -> Optional[str]:
    return os.getenv("CI_COMMIT_BEFORE_SHA")


def gitlab_merge_request_previous_commit_sha(verbose: bool) -> Optional[str]:
    targeted_branch = os.getenv("CI_MERGE_REQUEST_TARGET_BRANCH_NAME")

    # Not in a pull request workflow
    if targeted_branch is None:
        return None

    # Requires to fetch targeted branch to access current HEAD
    git(["fetch", "origin", targeted_branch])

    try:
        last_commit = get_last_commit_sha_of_branch(f"origin/{targeted_branch}")
    except Exception:
        # If fail to get last commit of target branch, fallback on CI env variable
        # "CI_MERGE_REQUEST_DIFF_BASE_SHA"
        # This is not the current state of the target branch but the initial state
        # of current branch
        if verbose:
            click.echo(f"Failed to get {targeted_branch} HEAD.")
            click.echo(
                f"Fallback on commit {os.getenv('CI_MERGE_REQUEST_DIFF_BASE_SHA')}"
            )
        return os.getenv("CI_MERGE_REQUEST_DIFF_BASE_SHA")

    return last_commit


PREVIOUS_COMMIT_SHA_FUNCTIONS = {
    SupportedCI.GITLAB: gitlab_previous_commit_sha,
    SupportedCI.GITHUB: github_previous_commit_sha,
}



from typing import Any, Tuple

import click

from ggshield.cmd.utils.common_options import add_common_options
from ggshield.core.config import Config

from .constants import DATETIME_FORMAT, FIELDS


@click.command()
@click.pass_context
@add_common_options()
def config_list_cmd(ctx: click.Context, **kwargs: Any) -> int:
    """
    Print the list of configuration keys and values.
    """
    config: Config = ctx.obj["config"]
    default_token_lifetime = config.auth_config.default_token_lifetime

    message_lines = []

    def add_entries(*entries: Tuple[str, Any]):
        for key, value in entries:
            message_lines.append(f"{key}: {value}")

    # List global values
    for field in FIELDS.values():
        config_obj = config.auth_config if field.auth_config else config.user_config
        value = getattr(config_obj, field.name)
        add_entries((field.name, value))
    message_lines.append("")

    # List instance values
    for instance in config.auth_config.instances:
        instance_name = instance.name or instance.url

        if instance.account is not None:
            workspace_id = instance.account.workspace_id
            token = instance.account.token
            token_name = instance.account.token_name
            expire_at = instance.account.expire_at
            expiry = (
                expire_at.strftime(DATETIME_FORMAT)
                if expire_at is not None
                else "never"
            )
        else:
            workspace_id = token = token_name = expiry = "not set"  # type: ignore

        _default_token_lifetime = (
            instance.default_token_lifetime
            if instance.default_token_lifetime is not None
            else default_token_lifetime
        )

        message_lines.append(f"[{instance_name}]")
        add_entries(
            ("default_token_lifetime", _default_token_lifetime),
            ("workspace_id", workspace_id),
            ("url", instance.url),
            ("token", token),
            ("token_name", token_name),
            ("expiry", expiry),
        )
        message_lines.append("")

    click.echo("\n".join(message_lines).strip())
    return 0
